# Copyright (C) 2011 W. Trevor King <wking@drexel.edu>
#
# This file is part of unfold_protein.
#
# Unfold_protein is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Unfold_protein is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with unfold_protein.  If not, see
# <http://www.gnu.org/licenses/>.

"""Define `UnfoldScanner` for sequential unfolding experiments."""

import signal as _signal

from calibcant.calibrate import move_far_from_surface as _move_far_from_surface
import pypiezo.base as _pypiezo_base

from . import LOG as _LOG
from .unfolder import ExceptionTooFar as _ExceptionTooFar
from .unfolder import ExceptionTooClose as _ExceptionTooClose

class UnfoldScanner (object):
    def __init__(self, config, unfolder):
        self.config = config
        self.unfolder = unfolder
        self._state = {'x direction': 1}

    def run(self):
        self._stop = False
        _signal.signal(_signal.SIGTERM, self._handle_stop_signal)
        for i in range(self.config['velocity']['num loops']):
            for velocity in self.config['velocity']['unfolding velocities']:
                if self._stop:
                    return
                self.unfolder.config['unfold']['velocity'] = velocity
                try:
                    self.unfolder.run()
                except _ExceptionTooFar:
                    self.stepper_approach()
                except _ExceptionTooClose:
                    self.move_far_from_surface()
                    self.stepper_approach()
                else:
                    self.position_scan_step()

    def _handle_stop_signal(self, signal, frame):
        self._stop = True

    def move_far_from_surface(self):
        _LOG.info('retract with the stepper motor by {} m'.format(
                self.unfolder.config['approach']['far']))
        _move_far_from_surface(
            stepper=self.unfolder.afm.stepper,
            distance=self.unfolder.config['approach']['far'])

    def stepper_approach(self):
        config = self.unfolder.config['approach']
        deflection = self.unfolder.read_deflection()
        setpoint = deflection + config['relative setpoint']
        def_config = self.unfolder.afm.piezo.config.select_config(
            'inputs', 'deflection')
        setpoint_bits = _pypiezo_base.convert_volts_to_bits(
            def_config, setpoint)
        self.unfolder.afm.stepper_approach(target_deflection=setpoint_bits)

    def position_scan_step(self):
        axis_name = 'x'
        config = self.config['position'] 
        axis_config = self.unfolder.afm.piezo.config.select_config(
                'axes', self.unfolder.afm.axis_name,
                get_attribute=_pypiezo_base.get_axis_name
                )
        pos = self.unfolder.afm.piezo.last_output[axis_name]
        pos_m = _pypiezo_base.convert_bits_to_meters(axis_config, pos)
        next_pos_m = pos_m + self._state['x direction']*config['x step']
        if next_pos_m > config['x max']:
            self._state['x direction'] = -1
            next_pos_m = pos_m + self._state['x direction']*config['x step']
        elif next_pos_m < config['x min']:
            self._state['x direction'] = 1
            next_pos_m = pos_m + self._state['x direction']*config['x step']
        next_pos = _pypiezo_base.convert_meters_to_bits(
            axis_config, next_pos_m)
        _LOG.info('move {} from {:g} to {:g} bits'.format(
                axis_name, pos, next_pos))
        self.unfolder.afm.piezo.jump(axis_name, next_pos)
